/*
    XenoProbes - a program to optimize probe placement in Xenoblade Chronicles X
    Copyright (C) 2015  Daniel K. O.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <fstream>
#include <string>
#include <memory>
#include <map>
#include <vector>
#include <queue>
#include <utility>
#include <cstdio>
#include <cstddef>
#include <algorithm>
#include <random>
#include <utility>
#include <iomanip>
#include <atomic>
#include <csignal>
#include <cassert>
#include <cmath>
#include <cstdint>

#include <sys/types.h>
#include <unistd.h>


#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/trim.hpp>
#include <boost/program_options.hpp>
#include <boost/filesystem.hpp>


#ifdef __MINGW32__
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <wincrypt.h>
#endif

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif



using std::size_t;
using std::cout;
using std::clog;
using std::cerr;
using std::endl;
using std::uint16_t;
using std::vector;
using std::string;


enum RunMode {
    HelpMode,
    PrintMode,
    OptimizeMode
};


float storageWeight;
float revenueWeight;
float productionWeight;


float mutationRate;
float eliteRatio;
size_t maxPopSize;
size_t tournamentRank;
size_t numOffsprings;
size_t maxIterations;
int maxAge;

string setupInput;
string inventoryInput;


bool piping = false;



#ifdef __MINGW32__
std::mt19937 mt; // note it wasn't seeded properly
#else
std::mt19937_64 mt{std::random_device{}()};
#endif


const vector<string> ores = {
    "Arc Sand Ore",
    "Aurorite",
    "Boiled-Egg Ore",
    "Bonjelium",
    "Cimmerian Cinnabar",
    "Dawnstone",
    "Enduron Lead",
    "Everfreeze Ore",
    "Foucaultium",
    "Infernium",
    "Lionbone Bort",
    "Marine Rutile",
    "Ouroboros Crystal",
    "Parhelion Platinum",
    "White Cometite"
};


int prodGradeToNum(char grade)
{
    switch (grade) {
        case 'A':
            return 500;
        case 'B':
            return 350;
        case 'C':
            return 250;
        default:
            return 0;
    }
}

int revGradeToNum(char grade)
{
    switch (grade) {
        case 'S':
            return 850;
        case 'A':
            return 750;
        case 'B':
            return 650;
        case 'C':
            return 550;
        case 'D':
            return 450;
        case 'E':
            return 300;
        case 'F':
            return 200;
        default:
            return 0;
    }
}


int combGradeToNum(char /*grade*/)
{
    return 0; // we don't know nor need this
}



struct Site {
    int name;
    int production = 0;
    int revenue = 0;
    int combat = 0;
    int sightseeing = 0;
    vector<uint16_t> ores;
    vector<uint16_t> neighbors;

    Site(int name, char production, char revenue,
         char combat, int sightseeing,
         const vector<uint16_t>& ores={},
         const vector<uint16_t>& neighbors={}) :
        name{name},
        production{prodGradeToNum(production)},
        revenue{revGradeToNum(revenue)},
        combat{combGradeToNum(combat)},
        sightseeing{sightseeing},
        ores{ores},
        neighbors{neighbors}
    {}
};


vector<Site> sites = {
    {101,'C','S','S',1},
    {102,'C','F','B',0},
    {103,'C','E','A',1},
    {104,'C','S','B',1},
    {105,'A','F','B',0},
    {106,'B','E','B',1,{0}},
    {107,'A','F','B',0},
    {108,'C','F','B',0,{1,0,8}},
    {109,'C','D','B',0,{8,5,10}},
    {110,'C','E','B',1,{1,0,14,5}},
    {111,'C','F','B',0,{8}},
    {112,'A','F','A',0},
    {113,'C','C','B',0},
    {114,'C','E','B',0},
    {115,'C','D','B',0,{0,14,10}},
    {116,'A','D','B',0},
    {117,'A','D','A',1},
    {118,'C','E','B',0,{1,14,5}},
    {119,'C','E','B',0},
    {120,'B','B','B',0},
    {121,'A','E','B',0},

    {201,'C','B','S',0},
    {202,'C','C','B',0,{4,7}},
    {203,'C','A','B',0,{4}},
    {204,'A','C','B',0},
    {205,'A','F','B',0},
    {206,'B','A','S',0},
    {207,'C','C','B',0,{9,14,4,8}},
    {208,'B','D','B',0,{8}},
    {209,'C','F','B',0},
    {210,'B','D','B',0},
    {211,'A','D','B',0},
    {212,'B','E','B',0,{1,6,14}},
    {213,'C','S','B',1},
    {214,'C','D','B',2},
    {215,'C','D','B',0,{1,6,7}},
    {216,'C','A','A',1},
    {217,'C','C','B',0,{1,9,4}},
    {218,'C','E','B',0,{1,6,14}},
    {219,'C','E','B',0,{6,14}},
    {220,'C','C','A',1,{9,7}},
    {221,'C','E','B',2},
    {222,'C','D','B',1},
    {223,'C','F','B',1},
    {224,'C','A','B',0},
    {225,'C','A','B',1},

    {301,'B','D','B',0,{9,0,10}},
    {302,'C','E','B',0},
    {303,'C','E','B',0,{1,14}},
    {304,'B','A','S',0},
    {305,'C','E','B',0,{1,0,6}},
    {306,'C','D','B',1},
    {307,'C','B','B',0,{9,0,6,14}},
    {308,'B','C','A',0,{12}},
    {309,'C','C','B',0,{6,12}},
    {310,'C','A','B',0},
    {311,'C','B','B',0},
    {312,'C','D','B',0,{9,2,10}},
    {313,'C','E','A',2},
    {314,'C','B','S',0},
    {315,'A','S','B',2},
    {316,'C','D','B',0},
    {317,'C','A','B',1},
    {318,'C','B','B',2,{2,14,10}},
    {319,'C','D','B',1,{9,2}},
    {320,'C','B','B',0,{1,12}},
    {321,'A','D','A',0},
    {322,'A','A','B',0},

    {401,'C','B','B',0,{13,11}},
    {402,'A','B','B',0},
    {403,'A','C','S',0},
    {404,'B','S','S',1},
    {405,'A','E','A',0,{0}},
    {406,'C','B','B',0},
    {407,'A','B','B',0},
    {408,'B','D','B',1,{1,0,7}},
    {409,'B','S','B',0},
    {410,'C','S','B',1},
    {411,'A','A','S',0},
    {412,'A','B','A',0},
    {413,'C','A','B',1},
    {414,'C','B','B',2,{13,11}},
    {415,'C','S','B',0},
    {416,'C','B','B',0},
    {417,'B','D','B',0,{7,2}},
    {418,'C','C','B',0,{13,0,7,2,11}},
    {419,'C','S','S',1},
    {420,'B','C','B',0,{7}},

    {501,'B','F','B',0,{0}},
    {502,'A','C','B',1,{3}},
    {503,'C','D','B',1,{6}},
    {504,'C','C','B',0,{0,6,11,3}},
    {505,'C','B','B',2},
    {506,'C','B','B',1,{3,0}},
    {507,'C','A','B',1,{3}},
    {508,'A','B','S',1,{6,11}},
    {509,'A','A','A',0},
    {510,'C','B','B',0,{3}},
    {511,'A','C','A',0,{3}},
    {512,'C','A','S',0},
    {513,'C','A','B',2},
    {514,'C','A','B',1},
    {515,'C','B','S',0},
    {516,'B','E','B',0}
};



enum ProbeType {
    Basic     = 0,
    Mining1   = 1,
    Mining2   = 2,
    Mining3   = 3,
    Mining4   = 4,
    Mining5   = 5,
    Mining6   = 6,
    Mining7   = 7,
    Mining8   = 8,
    Mining9   = 9,
    Mining10  = 10,
    Research1 = 11,
    Research2 = 12,
    Research3 = 13,
    Research4 = 14,
    Research5 = 15,
    Research6 = 16,
    Booster1  = 17,
    Booster2  = 18,
    Dupe      = 19,
    Storage   = 20,
    Invalid   = 21
};


using ProbeVec = vector<ProbeType>;


struct Solution {
    double score = 0;
    int age = 0;
    ProbeVec setup;
};


ProbeType str2ProbeType(const string& s)
{
    if (s == "B")
        return Basic;
    if (s[0] == 'M')
        return ProbeType(int(Mining1) + std::stoi(s.substr(1)) - 1);
    if (s[0] == 'R')
        return ProbeType(int(Research1) + std::stoi(s.substr(1)) - 1);
    if (s == "B1")
        return Booster1;
    if (s == "B2")
        return Booster2;
    if (s == "D")
        return Dupe;
    if (s == "S")
        return Storage;
    throw std::exception{};
}


string probeType2str(ProbeType type)
{
    switch (type) {
        case Basic:
            return "B";
        case Mining1:
        case Mining2:
        case Mining3:
        case Mining4:
        case Mining5:
        case Mining6:
        case Mining7:
        case Mining8:
        case Mining9:
        case Mining10:
            return "M" + std::to_string(type - Mining1 + 1);
        case Research1:
        case Research2:
        case Research3:
        case Research4:
        case Research5:
        case Research6:
            return "R" + std::to_string(type - Research1 + 1);
        case Booster1:
            return "B1";
        case Booster2:
            return "B2";
        case Dupe:
            return "D";
        case Storage:
            return "S";
        default:
            return "invalid";
    }
}

string probeType2longstr(ProbeType type)
{
    switch (type) {
        case Basic:
            return "Basic";
        case Mining1:
            return "Mining G1";
        case Mining2:
            return "Mining G2";
        case Mining3:
            return "Mining G3";
        case Mining4:
            return "Mining G4";
        case Mining5:
            return "Mining G5";
        case Mining6:
            return "Mining G6";
        case Mining7:
            return "Mining G7";
        case Mining8:
            return "Mining G8";
        case Mining9:
            return "Mining G9";
        case Mining10:
            return "Mining G10";
        case Research1:
            return "Research G1";
        case Research2:
            return "Research G2";
        case Research3:
            return "Research G3";
        case Research4:
            return "Research G4";
        case Research5:
            return "Research G5";
        case Research6:
            return "Research G6";
        case Booster1:
            return "Booster G1";
        case Booster2:
            return "Booster G2";
        case Dupe:
            return "Duplicator";
        case Storage:
            return "Storage";
        default:
            return "invalid";
    }
}





bool isSiteOp(const std::string& arg)
{
    if (arg.empty())
        return false;
    return arg[0] == '-' || arg[0] == '+';
}





struct Probe {
    ProbeType type;
    double production;
    double revenue;
    double boost;

    bool isBasic() const noexcept
    {
        return type == Basic;
    }

    bool isMining() const noexcept
    {
        return type >= Mining1 && type <= Mining10;
    }

    bool isResearch() const noexcept
    {
        return type >= Research1 && type <= Research6;
    }
    
    bool isBooster() const noexcept
    {
        return type >= Booster1 && type <= Booster2;
    }

    bool isDupe() const noexcept
    {
        return type == Dupe;
    }

    bool isStorage() const noexcept
    {
        return type == Storage;
    }
};

vector<Probe> probes = {
    {Basic     , 0.5, 0.5, 1.0},
    {Mining1   , 1.0, 0.3, 1.0},
    {Mining2   , 1.2, 0.3, 1.0},
    {Mining3   , 1.4, 0.3, 1.0},
    {Mining4   , 1.6, 0.3, 1.0},
    {Mining5   , 1.8, 0.3, 1.0},
    {Mining6   , 2.0, 0.3, 1.0},
    {Mining7   , 2.2, 0.3, 1.0},
    {Mining8   , 2.4, 0.3, 1.0},
    {Mining9   , 2.7, 0.3, 1.0},
    {Mining10  , 3.0, 0.3, 1.0},
    {Research1 , 0.3, 2.0, 1.0},
    {Research2 , 0.3, 2.5, 1.0},
    {Research3 , 0.3, 3.0, 1.0},
    {Research4 , 0.3, 3.5, 1.0},
    {Research5 , 0.3, 4.0, 1.0},
    {Research6 , 0.3, 4.5, 1.0},
    {Booster1  , 0.1, 0.1, 1.5},
    {Booster2  , 0.1, 0.1, 2.0},
    {Dupe      , 0.0, 0.0, 1.0},
    {Storage   , 0.1, 0.1, 1.0}
};

vector<ProbeType> inventory;



bool operator==(const Solution& a, const Solution& b)
{
    const auto sz = sites.size();
    return a.score == b.score &&
        std::equal(a.setup.begin(), a.setup.begin()+sz,
                   b.setup.begin());
}


bool operator<(const Solution& a, const Solution& b)
{
    return a.score < b.score;
}


bool operator>(const Solution& a, const Solution& b)
{
    return a.score > b.score;
}


vector<vector<string>>
loadCSV(const string& fname)
{
    vector<vector<string>> result;

    std::ifstream file{fname};
    if (!file)
        throw std::runtime_error{"could not open file " + fname};
    string line;
    while (getline(file, line)) {
        auto commentPos = line.find('#');
        if (commentPos != string::npos)
            line.erase(commentPos);
        boost::trim(line);
        if (line.empty())
            continue;
        vector<string> records;
        boost::split(records, line, [](char c) {return c==',';});
        result.push_back(std::move(records));
    }

    return result;
}


int findSiteIdx(int name)
{
    for (size_t i=0; i<sites.size(); ++i)
        if (sites[i].name == name)
            return i;
    return -1;
}


void tryConnect(int siteU, int siteV)
{
    int uidx = findSiteIdx(siteU);
    if (uidx == -1)
        return;
    int vidx = findSiteIdx(siteV);
    if (vidx == -1)
        return;
    sites[uidx].neighbors.push_back(vidx);
    sites[vidx].neighbors.push_back(uidx);
}


void connectSites()
{
    tryConnect(101,105);
    tryConnect(102,104);
    tryConnect(103,106);
    tryConnect(103,105);
    tryConnect(104,106);
    tryConnect(105,109);
    tryConnect(106,107);
    tryConnect(107,110);
    tryConnect(108,109);
    tryConnect(110,111);
    tryConnect(110,112);
    tryConnect(111,113);
    tryConnect(112,114);
    tryConnect(112,115);
    tryConnect(114,116);
    tryConnect(116,117);
    tryConnect(117,118);
    tryConnect(117,120);
    tryConnect(118,121);
    tryConnect(119,120);

    tryConnect(201,206);
    tryConnect(202,207);
    tryConnect(202,203);
    tryConnect(202,208);
    tryConnect(203,204);
    tryConnect(204,205);
    tryConnect(204,211);
    tryConnect(204,212);
    tryConnect(205,209);
    tryConnect(206,207);
    tryConnect(206,213);
    tryConnect(210,211);
    tryConnect(212,216);
    tryConnect(214,215);
    tryConnect(215,218);
    tryConnect(216,218);
    tryConnect(216,225);
    tryConnect(217,222);
    tryConnect(218,224);
    tryConnect(219,220);
    tryConnect(220,221);
    tryConnect(220,225);
    tryConnect(221,222);
    tryConnect(223,224);

    tryConnect(301,302);
    tryConnect(301,303);
    tryConnect(303,306);
    tryConnect(304,305);
    tryConnect(304,306);
    tryConnect(304,309);
    tryConnect(305,308);
    tryConnect(306,307);
    tryConnect(307,313);
    tryConnect(309,311);
    tryConnect(310,311);
    tryConnect(312,315);
    tryConnect(312,313);
    tryConnect(313,314);
    tryConnect(315,316);
    tryConnect(315,318);
    tryConnect(315,321);
    tryConnect(317,318);
    tryConnect(317,319);
    tryConnect(320,321);
    tryConnect(321,322);

    tryConnect(401,402);
    tryConnect(401,404);
    tryConnect(402,408);
    tryConnect(403,405);
    tryConnect(404,407);
    tryConnect(405,408);
    tryConnect(405,409);
    tryConnect(406,408);
    tryConnect(407,412);
    tryConnect(408,413);
    tryConnect(409,411);
    tryConnect(410,412);
    tryConnect(411,414);
    tryConnect(412,415);
    tryConnect(413,416);
    tryConnect(416,418);
    tryConnect(416,419);
    tryConnect(417,419);
    tryConnect(419,420);

    tryConnect(501,502);
    tryConnect(502,503);
    tryConnect(503,504);
    tryConnect(504,508);
    tryConnect(505,506);
    tryConnect(505,509);
    tryConnect(507,508);
    tryConnect(508,509);
    tryConnect(508,511);
    tryConnect(509,510);
    tryConnect(509,513);
    tryConnect(511,512);
    tryConnect(511,514);
    tryConnect(513,516);
    tryConnect(514,515);

    tryConnect(103,222);
    tryConnect(113,409);
    tryConnect(121,301);
    tryConnect(415,502);
}


ProbeVec loadSetup(const string& name)
{
    ProbeVec setup(sites.size());

    auto data = loadCSV(name);
    for (auto& record : data) {
        int siteName = std::stoi(record[0]);
        int siteIdx = findSiteIdx(siteName);
        if (siteIdx == -1)
            throw std::runtime_error{"Setup file uses a non-available site: " + std::to_string(siteIdx)};
        setup[siteIdx] = str2ProbeType(record[1]);
    }
    return setup;
}


int getComboSize(size_t source, const ProbeVec& setup)
{
    const auto stype = setup[source];

    int comboSize = 0;

    static vector<char> visited;
    static std::queue<size_t> Q;

    visited.clear();
    visited.resize(setup.size());

    Q.push(source);
    while (!Q.empty()) {
        size_t idx = Q.front();
        Q.pop();

        visited[idx] = true;
        ++comboSize;

        for (auto nidx : sites[idx].neighbors)
            if (setup[nidx] == stype && !visited[nidx])
                Q.push(nidx);

    }
    return comboSize;
}


double getComboBonus(size_t idx, const ProbeVec& setup)
{
    const auto& probe = probes[setup[idx]];
    if (probe.isDupe() || probe.isBooster())
        return 1.0;

    int sz = getComboSize(idx, setup);
    if (sz < 3)
        return 1.0;

    if (sz < 5)
        return 1.3;
    if (sz < 8)
        return 1.5;
    return 1.8;
}


double getProbeBoost(size_t idx, const ProbeVec& setup) noexcept
{
    const auto& probe = probes[setup[idx]];
    if (!probe.isDupe()) {
        if (probe.isBooster())
            return probe.boost * getComboBonus(idx, setup);
        else
            return probe.boost;
    }

    double boost = 1.0;
    for (auto nidx: sites[idx].neighbors) {
        const auto& neighbor = probes[setup[nidx]];
        boost *= neighbor.boost;
    }
    return boost;
}



double getProbeProduction(size_t idx, const ProbeVec& setup) noexcept
{
    auto type = setup[idx];
    
    const auto& probe = probes[type];
    const auto& site = sites[idx];

    double probeRate = probe.production;

    // check if it's in a combo
    if (probe.isMining())
        probeRate *= getComboBonus(idx, setup);

    double dupeStash = 0; // things that don't get boosted
    if (probe.isDupe()) {
        for (auto nidx : site.neighbors) {
            const auto& neighbor = probes[setup[nidx]];
            if (neighbor.isDupe())
                continue;
            if (neighbor.isMining())
                probeRate += neighbor.production;
            else
                dupeStash += neighbor.production;
        }
    }

    // check if there are neighbor boosters
    if (probe.isMining() || probe.isDupe()) {
        double scale = 1;
        for (auto nidx : site.neighbors) {
            //const auto& neighbor = probes[setup[nidx]];
            double boost = getProbeBoost(nidx, setup);
            scale *= boost;
        }
        //clog << site.name << " mining got boosted by " << scale << endl;
        probeRate *= scale;
    }

    return probeRate + dupeStash;
}


double getProduction(size_t idx, const ProbeVec& setup) noexcept
{
    return getProbeProduction(idx, setup) * sites[idx].production;
}



double getProbeRevenue(size_t idx, const ProbeVec& setup) noexcept
{
    auto type = setup[idx];
    
    const auto& probe = probes[type];
    const auto& site = sites[idx];

    double probeRate = probe.revenue;

    // check if it's in a combo
    if (probe.isResearch())
        probeRate *= getComboBonus(idx, setup);

    double dupeStash = 0; // things that don't get boosted
    if (probe.isDupe()) {
        for (auto nidx : site.neighbors) {
            const auto& neighbor = probes[setup[nidx]];
            if (neighbor.isDupe())
                continue;
            if (neighbor.isResearch())
                probeRate += neighbor.revenue;
            else
                dupeStash += neighbor.revenue;
        }
    }

    // check if there are neighbor boosters
    if (probe.isResearch() || probe.isDupe()) {
        double scale = 1;
        for (auto nidx : site.neighbors) {
            double boost = getProbeBoost(nidx, setup);
            scale *= boost;
        }
        //clog << site.name << " revenue got boosted by " << scale << endl;
        probeRate *= scale;
    }

    return probeRate + dupeStash;
}



double getRevenue(size_t idx, const ProbeVec& setup) noexcept
{
    const auto& site = sites[idx];

    double siteRev = site.revenue;
    if (probes[setup[idx]].isResearch())
        siteRev += 1000*site.sightseeing;

    return getProbeRevenue(idx, setup) * siteRev;
}


double getStorage(size_t idx, const ProbeVec& setup) noexcept
{
    auto type = setup[idx];
    
    const auto& probe = probes[type];
    const auto& site = sites[idx];

    double probeStorage = 0;

    // check if it's in a combo
    if (probe.isStorage()) {
        probeStorage = 3000;
        probeStorage *= getComboBonus(idx, setup);
    }

    if (probe.isDupe()) {
        for (auto nidx : site.neighbors) {
            const auto& neighbor = probes[setup[nidx]];
            if (neighbor.isStorage())
                probeStorage += 3000;
        }
    }

    // check if there are neighbor boosters
    if (probe.isStorage() || probe.isDupe()) {
        double scale = 1;
        for (auto nidx : site.neighbors) {
            //const auto& neighbor = probes[setup[nidx]];
            double boost = getProbeBoost(nidx, setup);
            scale *= boost;
        }
        //clog << site.name << " storage got boosted by " << scale << endl;
        probeStorage *= scale;
    }

    return probeStorage;
}




bool killSite(int name)
{
    auto idx = findSiteIdx(name);
    if (idx == -1) {
        cerr << "could not remove the site " << name << endl;
        return false;
    }
    sites.erase(sites.begin()+idx);
    return true;
}


bool handleSiteUpdate(const vector<string>& record)
{
    // player doesn't have full map
    char op;
    int siteName;
    if (sscanf(record[0].c_str(), "%c%d", &op, &siteName) != 2) {
        cerr << "error parsing site update: " << record[0] << endl;
        return false;
    }
    
    if (op != '+' && op != '-') {
        cerr << "invalid site operation: must be either + or -" << endl;
        return false;
    }
    
    switch (op) {
        case '+': {
            // changing the sightseeing spots
            int num = std::stoi(record[1]);
            auto idx = findSiteIdx(siteName);
            if (idx == -1) {
                cerr << "could not update sightseeing of site " << siteName << endl;
                return false;
            }
                
            sites[idx].sightseeing = num;
            clog << "updated sightseeing spots for site "
                 << siteName << "  to " << num << endl;
            return true;
        }
        case '-':
            // player got no probe on this site
            return killSite(siteName);
    }
    return false;
}


void loadInventory(const string& fileName)
{
    inventory.clear();

    int modifiedSites = 0;
    
    try {
        auto data = loadCSV(fileName);
        for (const auto& record : data) {
            if (isSiteOp(record[0])) {
                if (handleSiteUpdate(record))
                    ++modifiedSites;
            } else {
                // just probe inventory data
                int num = std::stoi(record[1]);
                for (int i=0; i<num; ++i)
                    inventory.push_back(str2ProbeType(record[0]));
            }
        }

        // ensure inventory is as big as needed to fill the entire map
        while (inventory.size() < sites.size())
            inventory.push_back(Basic);

        clog << "Inventory has " << inventory.size() << " probes";
        auto numBasic = std::count(inventory.begin(), inventory.end(), Basic);
        if (numBasic > 0)
            clog << ", and " << numBasic << " are Basic Probes." << endl;
        else
            clog << "." << endl;

        if (modifiedSites)
            clog << "Sites modified by the inventory: " << modifiedSites << endl;
    }
    catch (std::exception& e) {
        throw std::runtime_error{"Loading " + fileName + string(" inventory: ") + e.what()};
    }
}


void printInventory()
{
    std::map<ProbeType, int> histo;
    for (auto item : inventory)
        ++histo[item];
    for (auto& entry : histo)
        if (entry.first != Basic)
            cout << probeType2str(entry.first) << "," << entry.second << endl;
}


ProbeVec getRandomSetup()
{
    ProbeVec setup = inventory;    
    std::shuffle(setup.begin(), setup.end(), mt);
    return setup;
}


void printSetup(const ProbeVec& setup)
{
    if (!piping) {
        int prevName = 100;
        for (size_t idx=0; idx<sites.size(); ++idx) {
            if (setup[idx] != Basic) {
                if (sites[idx].name / 100 != prevName / 100)
                    cout << endl;
                cout << sites[idx].name << "," << probeType2str(setup[idx]) << endl;
                prevName = sites[idx].name;
            }
        }
    } else {
        cout << endl << "Probe configuration:" << endl;
        int prevName = 100;
        for (size_t idx=0; idx<sites.size(); ++idx) {
            if (setup[idx] != Basic) {
                if (sites[idx].name / 100 != prevName / 100)
                    cout << endl;
                cout << "FN " << sites[idx].name << " : " << probeType2longstr(setup[idx]) << endl;
                prevName = sites[idx].name;
            }
        }
    }
}


void mutate(vector<ProbeType>& setup, double prob) noexcept
{
    assert(!setup.empty());
    assert(setup.size() >= sites.size());
    std::bernoulli_distribution dist{prob};
    // any swap is equally likely, but we don't care about extra probes in the inventory
    for (size_t idx=0; idx<sites.size()-1; ++idx) {
        if (dist(mt)) {
            std::uniform_int_distribution<size_t> idist{idx, setup.size()-1};
            size_t dest = idist(mt);
            std::swap(setup[idx], setup[dest]);
        }
    }
}


template<class T>
T randomElement(const vector<T>& bag)
{
    assert(!bag.empty());
    std::uniform_int_distribution<size_t> dist{0, bag.size()-1};
    return bag[dist(mt)];
}


float evaluate(ProbeVec& setup) noexcept
{
    long totalProd = 0;
    long totalRev = 0;
    long totalStorage = 6000;
    for (size_t i = 0; i<sites.size(); ++i) {
        totalProd += getProduction(i, setup);
        totalRev += getRevenue(i, setup);
        totalStorage += getStorage(i, setup);
    }

    return totalStorage * storageWeight
        + totalRev * revenueWeight
        + totalProd * productionWeight;
}


void printTotals(const ProbeVec& setup)
{
    long totalProd = 0;
    long totalRev = 0;
    long totalStorage = 6000;
    // keep track of which ores have a mining probe on them
    // TODO: figure out how different probes affect ore extraction
    std::set<uint16_t> minedOres;
    for (size_t i = 0; i<sites.size(); ++i) {
        totalProd += getProduction(i, setup);
        totalRev += getRevenue(i, setup);
        totalStorage += getStorage(i, setup);

        const Probe& probe = probes[setup[i]];
        if (probe.isMining()) {
            for (auto& oreId : sites[i].ores)
                minedOres.insert(oreId);
        }
    }
    cout << "# Miranium: " << std::setw(6) << totalProd << "\n"
         << "# Revenue:  " << std::setw(6) << totalRev << "\n"
         << "# Storage:  " << std::setw(6) << totalStorage << "\n"
         << "# Ores (" << minedOres.size() << "/" << ores.size() << "):\n";
    vector<string> oreNames;
    for (const auto& oreId : minedOres)
        oreNames.push_back(ores[oreId]);
    std::sort(oreNames.begin(), oreNames.end());
    for (const auto& name : oreNames)
        cout << "#  " << name << endl;
}


bool areSame(const ProbeVec& a, const ProbeVec& b)
{
    const auto sz = sites.size();
    return std::equal(a.begin(), a.begin()+sz,
                      b.begin());
}


bool isValid(const ProbeVec& setup)
{
    std::map<ProbeType,int> histoInv;
    std::map<ProbeType,int> histoSetup;
    
    for (const auto& item : inventory)
        ++histoInv[item];
    for (const auto& item : setup)
        ++histoSetup[item];
    return histoInv == histoSetup;
}



std::atomic<bool> shouldStop;


void handleSIGINT(int)
{
    shouldStop = true;
}



RunMode parseOptions(int argc, char** argv)
{
    namespace po = boost::program_options;
    
    po::options_description desc("Allowed options");
    desc.add_options()
        ("help,h", "Prints help message.")
        ("print",
         po::value<string>(&setupInput),
         "Just print estimated output for a given setup file (in CSV format.)")
        ("storageweight",
         po::value<float>(&storageWeight)->default_value(1000),
         "How important is Storage (maximum Miranium that can be held at once.)")
        ("revenueweight",
         po::value<float>(&revenueWeight)->default_value(10),
         "How important is Revenue (cash output.)")
        ("productionweight",
         po::value<float>(&productionWeight)->default_value(1),
         "How important is Production (Miranium output.)")
        ("iterations",
         po::value<size_t>(&maxIterations)->default_value(2000),
         "How many iterations to run.")
        ("offsprings",
         po::value<size_t>(&numOffsprings)->default_value(100),
         "How many random variations to generate from each solution.")
        ("mutation,m",
         po::value<float>(&mutationRate)->default_value(1.0/32),
         "Mutation rate (probability of any probe swapping places with another.)")
        ("inventory,i",
         po::value<string>(&inventoryInput)->default_value("inventory.txt"),
         "The inventory file, in CSV format.")
        ("population",
         po::value<size_t>(&maxPopSize)->default_value(200),
         "Maximum population size.")
        ("age",
         po::value<int>(&maxAge)->default_value(50),
         "Maximum age for a stuck solution; after this many generations, replace this solution by a new, random one. Set to zero to disable discarding stuck solutions.")
        ("pipe", "Prints the process ID int he first line, and sends everything to standard output.")
        ;
    po::variables_map vm;
    po::store(po::parse_command_line(argc, argv, desc), vm);
    po::notify(vm);

    if (vm.count("help")) {
        cerr << desc << endl;
        return HelpMode;
    }

    if (vm.count("print"))
        return PrintMode;

    if (vm.count("pipe"))
        piping = true;
    
    return OptimizeMode;
}


void doHillClimbing()
{
    clog << "Starting Hill Climbing with parameters:" << endl;
    clog << "  storage weight   = " << std::setw(6) << storageWeight << endl
         << "  revenue weight   = " << std::setw(6) << revenueWeight << endl
         << "  production weight= " << std::setw(6) << productionWeight << endl;
    clog << "  iterations=" << maxIterations << "  offsprings=" << numOffsprings
         << "  mutation=" << mutationRate << "  age=" << maxAge
         << "  population=" << maxPopSize << endl;


    vector<Solution> population(maxPopSize), newGen;
    for (size_t i=0; i<maxPopSize; ++i) {
        population[i].setup = getRandomSetup();
        population[i].score = evaluate(population[i].setup);
    }

    Solution globalBest, best, worst;
    size_t killed = 0;
    for (size_t iter=0; iter<maxIterations && !shouldStop; ++iter) {

        clog << "iteration: " << (iter+1) << "/" << maxIterations
             << "  best: " << std::setprecision(9) << best.score
             << "  worst: " << std::setprecision(9) << worst.score
             << "  killed: " << killed
             << endl;

        newGen.clear();
        killed = 0;

        for (const auto& ancestor : population) {
            auto bestChild = ancestor;
            Solution child;
            for (size_t i=0; i<numOffsprings; ++i) {
                child = ancestor;
                child.age = 0;
                mutate(child.setup, mutationRate);

                if (areSame(ancestor.setup, child.setup))
                    continue;

                child.score = evaluate(child.setup);
                if (child > bestChild)
                    bestChild = std::move(child);
            }
            if (bestChild.score == ancestor.score && maxAge > 0) {
                if (++bestChild.age > maxAge) {
                    bestChild.setup = getRandomSetup();
                    bestChild.score = evaluate(bestChild.setup);
                    bestChild.age = 0;
                    ++killed;
                }
            }

            newGen.push_back(std::move(bestChild));
        }

        auto extremes = std::minmax_element(newGen.begin(), newGen.end());
        worst = *extremes.first;
        best = *extremes.second;

        if (best > globalBest)
            globalBest = best;

        swap(population, newGen);
    }
    if (shouldStop)
        clog << "\nOptimization stopped before completion.\n"; // break line with ^C character

    printSetup(globalBest.setup);
    printTotals(globalBest.setup);
    cout << "# Best score: " << std::setprecision(9) << globalBest.score << endl;
}



#ifdef __MINGW32__
void doWin32Init()
{
    // make sure the RNG is properly seeded
    {
        HCRYPTPROV provider = 0;

        if (!::CryptAcquireContextW(&provider, 0, 0, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT | CRYPT_SILENT))
            throw std::runtime_error{"could not acquire crypto context"};

        const DWORD length = 256;
        BYTE buffer[length];

        if (!::CryptGenRandom(provider, length, buffer)) {
            ::CryptReleaseContext(provider, 0);
            throw std::runtime_error{"could not generate random numbers!"};
        }

        std::seed_seq seq(buffer, buffer+length);
        mt.seed(seq);

        if (!::CryptReleaseContext(provider, 0))
            throw std::runtime_error{"could not release crypto context"};
    }


    {
        // set the current working directory
        namespace fs = boost::filesystem;

        vector<char> buffer(1024);
        int length = ::GetModuleFileNameA(nullptr, &buffer[0], buffer.size());
        if (::GetLastError() != ERROR_INSUFFICIENT_BUFFER) {
            string exeFilename(buffer.begin(), buffer.begin()+length);
            fs::path exePath = fs::path(exeFilename).parent_path();
            clog << "setting working directory to " << exePath << endl;
            fs::current_path(exePath);
        }
    }

}
#endif


int main(int argc, char** argv)
{
    clog << PACKAGE_STRING << endl;

#ifdef __MINGW32__
    doWin32Init();
#endif

    auto mode = parseOptions(argc, argv);

    if (mode == HelpMode)
        return 0;

    if (mode == PrintMode) {
        auto setup = loadSetup(setupInput);
        printSetup(setup);
        printTotals(setup);
        return 0;
    }

    if (piping) {
        cout << getpid() << endl;
        clog.rdbuf(cout.rdbuf());
        cerr.rdbuf(cout.rdbuf());
    }

    loadInventory(inventoryInput);

    // note: connect AFTER the sites were potentially deleted
    connectSites();

    std::signal(SIGINT, handleSIGINT);

    doHillClimbing();
}

